/** File nonlfit.c
 *
 * Adapted to GraceGTK by P. Vincent
 * 2012 : added LOESS interface
 *
 *
 * Grace - GRaphing, Advanced Computation and Exploration of data
 * Home page for Grace: http://plasma-gate.weizmann.ac.il/Grace/
 * Copyright (c) 1991-1995 Paul J Turner, Portland, OR
 * Copyright (c) 1996-2000 Grace Development Team
 * Grace is maintained by Evgeny Stambulchik
 * 
 * 
 *                           All Rights Reserved
 * 
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 * 
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 * 
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* 
 *
 * nonlinear curve fitting
 *
 */

#include <config.h>

#include <cmath.h>

#include "globals.h"
#include "graphs.h"
#include "utils.h"
#include "parser.h"
#include "protos.h"

#include "nonlfit.h"

/* Needed only for `integer' and `doublereal' definitions */
#include "f2c.h"

#define MAXPARM 10        /* max number of parameters for non-lin fit */
#define MAXFIT 12         /* max degree of polynomial+1 that can be fitted */

NONLparams 	nonl_parms[MAXPARM];
NONLopts   	nonl_opts;
NONL_load_prefs nonl_prefs;

#ifdef HAVE_BLAS
/* Must be synchronized with NL_family in ../loess/loess.h */
static char *family[]  = {"symmetric" ,"gaussian"};
static char *linquad[] = {"linear" ,"quadratic"};
#endif /* HAVE_BLAS */


extern int lmdif_(U_fp, integer *, integer *, doublereal *,
            doublereal *, doublereal *, doublereal *, doublereal *, 
	    integer *, doublereal *, doublereal *, integer *, doublereal *, 
	    integer *, integer *, integer *, doublereal *, integer *, integer 
	    *, doublereal *, doublereal *, doublereal *, doublereal *, 
	    doublereal *);

static double *xp, *yp, *y_saved;
static double *wts;
static char *ra;

static int src_gno  ,src_setno;
static int dest_gno ,dest_setno;


static int niter = 0;
static int trace_on = FALSE;

int lmdif_drv(U_fp fcn, integer m, integer n, doublereal *x, 
	      doublereal *fvec, doublereal *tol, integer *iwa, 
	      doublereal *wa, integer lwa, integer nsteps
	      ,integer nprint);

void a_to_parms (double *a);
void parms_to_a (double *a);

void nonlfit_set_defaults (void)
{
  int i;
  for (i = 0; i < MAXPARM; i++) {
    nonl_parms[i].value  = 1.0;
    nonl_parms[i].constr = FALSE;
    nonl_parms[i].min 	 = 0.0;
    nonl_parms[i].max 	 = 10.0;
    nonl_parms[i].step 	 = 0.1;
  }
  nonl_opts.nonl_method = NL_LEVENBERG;
  nonl_opts.title     	= copy_string (nonl_opts.title, "A fit");
  nonl_opts.rtype     	= RESTRICT_NONE;
  nonl_opts.rnegate   	= FALSE;
  nonl_opts.wmethod   	= NL_WEIGHT_NONE;
  nonl_opts.wfunc     	= NULL;
  nonl_opts.formula   	= copy_string (nonl_opts.formula, "y = ");
  nonl_opts.parnum    	= 2;
  nonl_opts.tolerance 	= 0.01;
  nonl_opts.nsteps    	= 5;
  nonl_opts.trace_on  	= FALSE;
  nonl_opts.logpts    	= FALSE;
  nonl_opts.family             = NL_GAUSSIAN;
  nonl_opts.span               =  0.75; /* Standard value use in original LOESS soft */
  nonl_opts.confidence_compute = FALSE; /* whether we compute confidence bars */
  nonl_opts.confidence_level   = 0.9;   /* the level used to compute confidence bars */
  nonl_opts.confidence_len     = 3;     /* the number of confidence bars to be computed */
  nonl_prefs.autoload 	= TRUE;
  nonl_prefs.load     	= NL_LOAD_VALUES;
  nonl_prefs.npoints  	= 10;
  nonl_prefs.start    	= 0.0;
  nonl_prefs.stop     	= 10.0;
  return;
}

void nonlfit_set_parm_value (int i ,double x)
{
  nonl_parms[i].value = x;
}

void initialize_nonl (void)
{
    nonl_opts.title   = NULL;
    nonl_opts.formula = NULL;
    nonlfit_set_defaults ();
}



void a_to_parms (double *a)
{
    int i;
    double t;
    
    for (i = 0; i < nonl_opts.parnum; i++) {
    	if (nonl_parms[i].constr) {
    	    /* map (-inf, inf) to (-1, 1) */
    	    t = a[i]/(abs(a[i]) + 1.0);
    	    
    	    /* map (-1, 1) to (nonl_lowb[i], nonl_uppb[i]) */
    	    nonl_parms[i].value = (nonl_parms[i].min + nonl_parms[i].max)/2.0 +
    	    	    (nonl_parms[i].max - nonl_parms[i].min)/2.0 * t;
    	} else {
    	    nonl_parms[i].value = a[i];
    	}
    }
}

/**
 *  nonlfit_compute_and_load: compute the values of the function in nonl_opts.formula
 *  and record the result into (dest_gno, set2) if(force || nonl_prefs.autoload)
 */
int nonlfit_compute_and_load (int gno1 ,int set1 ,int gno2 ,int set2 ,int force)
{
  int i, npts = 0;
  double delx, *xfit, *y, *yfit;
  if (nonl_prefs.load == NL_LOAD_FUNCTION) {
    if (nonl_prefs.npoints <= 1) {
      errmsg ("Number of points must be > 1");
      return RETURN_FAILURE;
    }
    src_gno    = gno1;
    src_setno  = set1;
    dest_gno   = gno2;
    dest_setno = set2;
    /* added djconnel Feb 2010 */
    if (nonl_opts.logpts &&
	(((nonl_prefs.start <= 0.0) && (nonl_prefs.stop >= 0.0)) ||
	 ((nonl_prefs.start >= 0.0) && (nonl_prefs.stop <= 0.0)))
	) {
      errmsg("Logarithmic points require same sign for start and stop");
      return RETURN_FAILURE;
    }
  }
  if (force || nonl_prefs.autoload) {
    switch (nonl_prefs.load) {
    case NL_LOAD_VALUES:
    case NL_LOAD_RESIDUALS:
      npts = getsetlength (src_gno ,set1);
      setlength (dest_gno ,set2 ,npts);
      copycol2 (src_gno ,set1 ,dest_gno ,set2 ,DATA_X);
      break;
    case NL_LOAD_FUNCTION:
      npts  = nonl_prefs.npoints;
      setlength (dest_gno ,set2 ,npts);
      delx = (nonl_prefs.stop - nonl_prefs.start)/(npts - 1);
      xfit = getx(dest_gno ,set2);
      for (i = 0; i < npts; i++) {
	/* logarithmic point progression added djconnel 10 Feb 2010 */
	if (nonl_opts.logpts) {
	  double f = ((double) i) / (npts - 1);
	  xfit[i] =
	    exp(
		log(fabs(nonl_prefs.start)) * (1.0 - f) +
		log(fabs(nonl_prefs.stop)) * f
		) * ((nonl_prefs.start > 0) ? 1 : -1);
	}
	else
	  xfit[i] =
	    nonl_prefs.start + i * delx;
      }
      break;
    }
    do_compute (dest_gno ,set2 ,dest_gno ,set2 ,NULL ,nonl_opts.formula);
    if (nonl_prefs.load == NL_LOAD_RESIDUALS) { /* load residuals */
      y    = gety (src_gno ,set1);
      yfit = gety (dest_gno ,set2);
      for (i = 0; i < npts; i++) {
	yfit[i] -= y[i];
      }
      comment_and_legend_set (dest_gno ,set2 ,"Levenberg fit residuals");
    } else {
      comment_and_legend_set (dest_gno ,set2 ,nonl_opts.formula);
    }
  }
  return RETURN_SUCCESS;
}


void parms_to_a (double *a)
{
    int i;
    double t;
        
    double eps  = 1.e-6;
       
    for (i = 0; i < nonl_opts.parnum; i++) {
    	if (nonl_parms[i].constr) {
    	    t = (nonl_parms[i].value - (nonl_parms[i].min + nonl_parms[i].max)/2.0)/
    	    	  ((nonl_parms[i].max - nonl_parms[i].min)/2.0);
    	    if (t < -(1.0 - eps)){
    	    	t = -(1.0 - eps);
    	    	nonl_parms[i].value = (nonl_parms[i].min + nonl_parms[i].max)/2.0 +
    	    	    (nonl_parms[i].max - nonl_parms[i].min)/2.0 * t;
    	    }
    	    if (t > (1.0 - eps)){
    	    	t = (1.0 - eps);
    	    	nonl_parms[i].value = (nonl_parms[i].min + nonl_parms[i].max)/2.0 +
    	    	    (nonl_parms[i].max - nonl_parms[i].min)/2.0 * t;
    	    }
    	    a[i] = t/(1.0 - abs(t)); 
    	} else {
    	    a[i] = nonl_parms[i].value;
    	}
    }
}

/* The approximating function, 
 * Called by do_nonlfit via lmdif_drv, at start, *iflag = 1
*/
void fcn (int * m, int * n, double * x, double * fvec,
		int * iflag)
{
    int errpos;
    int i;
    char itbuf[64];
    double *out_y;

    a_to_parms(x);

    /* apply formula */
    errpos = scanner(nonl_opts.formula);
    if (errpos) {
	errmsg("error in fcn");
	*iflag = -1;
	return;
    }

    /* if nprint > 0, trace is on and 
     *  we create a new set at each iteration (i.e. when *iflag == 0)
     */
    if (*iflag == 0) {                   /* it is a new iteration */
      niter++;
      int newset = nextset (dest_gno);
      do_copyset (src_gno ,src_setno ,dest_gno ,newset);
      out_y = gety (dest_gno ,newset);
      for (i = 0; i < *m; ++i) {
	out_y[i] = yp[i];
      }
      sprintf (itbuf ,"fit (G%d.%d, %d)" ,dest_gno ,dest_gno ,niter);
      comment_and_legend_set (dest_gno ,newset ,itbuf);
      sprintf (itbuf ," %d \t" ,niter);
      stufftext (itbuf);
      for (i = 0; i < *n; ++i) {
	sprintf (itbuf ,"a%d = %g\t" ,i ,x[i]);
	stufftext (itbuf);
      }
      sprintf(itbuf ,"\n");
      stufftext (itbuf);
    } else if (*iflag == 1 && trace_on == FALSE) {
      niter++;
    }

    /* compute residus (difference) */
    for (i = 0; i < *m; ++i) {
    	fvec[i] = yp[i] - y_saved[i];
    }
    /* apply weigh function, if any */
    if (wts != NULL) {
         for (i = 0; i < *m; ++i) {
             fvec[i] *= sqrt(fabs(wts[i]));
         }
    }
    /* apply restriction, if any */
    if (ra != NULL) {
         for (i = 0; i < *m; ++i) {
             fvec[i] *= ra[i];
         }
    }
}

/*
 * find correlation coefficient
 */
int correlation(double *x, double *y, int n, double *cor)
{
    double xbar, xsd;
    double ybar, ysd;
    int i;

    *cor = 0.0;
    if (n < 2) {
        return RETURN_FAILURE;
    }
    
    stasum(x, n, &xbar, &xsd);
    stasum(y, n, &ybar, &ysd);
    if (xsd == 0.0 || ysd == 0.0) {
        return RETURN_FAILURE;
    }
    
    for (i = 0; i < n; i++) {
        *cor += (x[i] - xbar)*(y[i] - ybar);
    }
    *cor /= ((n - 1)*xsd*ysd);
    
    return RETURN_SUCCESS;
}

/**
 * Levenberg method
 * Fit (gno,setno) using weigth array "warray" and region "rarray" with nteps maxi
 * If nprint==TRUE, create a new set at each iteration step (trace is on).
 * The computed coefficients are put in global nonl_parms
 */
int do_nonlfit(int gno, int setno, double *warray, char *rarray ,int gno2)
{
    int info = -1;
    double *fvec, *wa;
    int i, n;
    integer lwa, iwa[MAXPARM];
    double a[MAXPARM];
    int parnum = nonl_opts.parnum;
    char buf[MAX_STRING_LENGTH];
    double cor, chisq, rms_pe, ysq, theil;
    int rms_ok;

    if (set_parser_setno(gno, setno) != RETURN_SUCCESS) {
	return RETURN_FAILURE;
    }
    n = getsetlength(gno, setno);
    
    lwa = (integer) n * parnum + 5 * parnum + n;
        
    fvec = xcalloc(n, SIZEOF_DOUBLE);
    if (fvec == NULL) {
	return RETURN_FAILURE;
    }
      
    y_saved = xcalloc(n, SIZEOF_DOUBLE);
    if (y_saved == NULL) {
	xfree(fvec);
	return RETURN_FAILURE;
    }

    wa = xcalloc(lwa, sizeof(doublereal));
    if (wa == NULL) {
	xfree(y_saved);
	xfree(fvec);
	return RETURN_FAILURE;
    }
    /* to use for tracing in fcn */
    src_gno   = gno;
    src_setno = setno;
    dest_gno  = gno2;

    sprintf(buf, "\nFitting G%d.s%d with formula: " ,gno ,setno);
    stufftext(buf);
    stufftext(nonl_opts.formula);
    stufftext("\n");
    stufftext("Initial guesses:\n");
    for (i = 0; i < nonl_opts.parnum; i++) {
        sprintf(buf, "\ta%1d = %g\n", i, nonl_parms[i].value);
        stufftext(buf);
    }
    sprintf(buf, "Tolerance = %g\n", nonl_opts.tolerance);
    stufftext(buf);
    
    xp = getx(gno, setno);
    yp = gety(gno, setno);
    
    for (i = 0; i < n; ++i) {
    	y_saved[i] = yp[i];
    }
   
    ra = rarray;
    wts = warray;

    parms_to_a(a);
    
    info = lmdif_drv((U_fp) fcn, (integer) n, (integer) parnum, a, fvec, 
		     &nonl_opts.tolerance, iwa, wa, lwa, 
		     (integer) nonl_opts.nsteps , (integer) nonl_opts.trace_on);
    
    a_to_parms(a);
    
    correlation(yp, y_saved, n, &cor);
    
    chisq = 0.0;
    theil = 0.0;
    ysq = 0.0;
    rms_ok = TRUE;
    for (i = 0; i < n; ++i) {
    	chisq += (yp[i] - y_saved[i])*(yp[i] - y_saved[i]);
        ysq += (y_saved[i]*y_saved[i]);
    	theil += (yp[i] - y_saved[i])*(yp[i] - y_saved[i]);
    	if (y_saved[i] == 0.0) {
            rms_ok = FALSE;
        }
    }
    theil = sqrt(theil/ysq);

    rms_pe = 0.0;
    if (rms_ok) {
        for (i = 0; i < n; ++i) {
    	    rms_pe += (yp[i] - y_saved[i])*(yp[i] - y_saved[i])/
                (y_saved[i]*y_saved[i]);
        }
        rms_pe = sqrt(rms_pe/n);
    }

    for (i = 0; i < n; ++i) {
    	yp[i] = y_saved[i];
    }
    
    xfree(y_saved);
    xfree(fvec);
    xfree(wa);

    if (info >= 0 && info <= 7) {
        char *s;
        switch (info) {
        case 0:
            s = "Improper input parameters.\n";
            break;
        case 1:
            s = "Relative error in the sum of squares is at most tol.\n";
            break;
        case 2:
            s = "Relative error between A and the solution is at most tol.\n";
            break;
        case 3:
            s = "Relative error in the sum of squares and A and the solution is at most tol.\n";
            break;
        case 4:
            s = "Fvec is orthogonal to the columns of the jacobian to machine precision.\n";
            break;
        case 5:
            s = "\n";
            break;
        case 6:
            s = "Tol is too small. No further reduction in the sum of squares is possible.\n";
            break;
        case 7:
            s = "Tol is too small. No further improvement in the approximate solution A is possible.\n";
            break;
        default:
            s = "\n";
            errmsg("Internal error in do_nonlfit()");
            break;
        }
	stufftext (s); 	//053 warnmsg (s); 
    }
    
    if ((info > 0 && info < 4) || (info == 5)) {
      sprintf (buf ,"Computed values after %d steps:\n" ,niter);
      stufftext(buf);
        for (i = 0; i < nonl_opts.parnum; i++) {
            sprintf(buf, "\ta%1d = %g\n", i, nonl_parms[i].value);
            stufftext(buf);
        }
        stufftext("\n");
        sprintf(buf, "Chi-square: %g\n", chisq);
        stufftext(buf);
        sprintf(buf, "Correlation coefficient: %f\n", cor);
        stufftext(buf);
        if (rms_ok) {
            sprintf(buf, "RMS per cent error: %g\n", rms_pe);
            stufftext(buf);
        }
        sprintf(buf, "Theil U coefficent: %g\n", theil);
        stufftext(buf);
    }

    return RETURN_SUCCESS;
}

/**
 * Interface for do_nonlfit allowing to create a new set if needed
 */
int do_nonlfit_command (int gno1 ,int setno1 ,int gno2 ,int setno2 ,double *warray ,int nsteps ,int nprint)
{
  if (is_valid_gno (gno2)) {
    dest_gno  = gno2;
  } else {
    dest_gno = get_cg ();
  }
  if (is_valid_setno (dest_gno ,setno2)) {
    dest_setno = setno2;
  } else {
    dest_setno = nextset (dest_gno);
  }
  return do_nonlfit (gno1 ,setno1 ,warray ,NULL ,dest_gno);
}



/*
 *     lmdif1 - modified by E. Stambulchik  to suit ACE/gr needs
 *
 *     argonne national laboratory. minpack project. march 1980.
 *     burton s. garbow, kenneth e. hillstrom, jorge j. more
 */
int lmdif_drv(U_fp fcn, integer m, integer n, doublereal *x, 
	      doublereal *fvec, doublereal *tol, integer *iwa, 
	      doublereal *wa, integer lwa, integer nsteps 
	      ,integer nprint)
{
    /* Initialized data */

    static doublereal factor = 100.;
    static doublereal zero = 0.;

    static integer mp5n, mode, nfev;
    static doublereal ftol, gtol, xtol;
    static doublereal epsfcn;
    static integer maxfev;
    static integer info;

/*
 *       fcn is the name of the user-supplied subroutine which
 *         calculates the functions. fcn must be declared
 *         in an external statement in the user calling
 *         program, and should be written as follows.
 * 
 *         subroutine fcn(m,n,x,fvec,iflag)
 *         integer m,n,iflag
 *         double precision x(n),fvec(m)
 *         ----------
 *         calculate the functions at x and
 *         return this vector in fvec.
 *         ----------
 *         return
 *         end
 * 
 *         the value of iflag should not be changed by fcn unless
 *         the user wants to terminate execution of lmdif1.
 *         in this case set iflag to a negative integer.
 * 
 *       m is a positive integer input variable set to the number
 *         of functions.
 * 
 *       n is a positive integer input variable set to the number
 *         of variables. n must not exceed m.
 * 
 *       x is an array of length n. on input x must contain
 *         an initial estimate of the solution vector. on output x
 *         contains the final estimate of the solution vector.
 * 
 *       fvec is an output array of length m which contains
 *         the functions evaluated at the output x.
 * 
 *       tol is a nonnegative input variable. termination occurs
 *         when the algorithm estimates either that the relative
 *         error in the sum of squares is at most tol or that
 *         the relative error between x and the solution is at
 *         most tol.
 * 
 *       info is an integer output variable. if the user has
 *         terminated execution, info is set to the (negative)
 *         value of iflag. see description of fcn. otherwise,
 *         info is set as follows.
 * 
 *         info = 0  improper input parameters.
 * 
 *         info = 1  algorithm estimates that the relative error
 *                   in the sum of squares is at most tol.
 * 
 *         info = 2  algorithm estimates that the relative error
 *                   between x and the solution is at most tol.
 * 
 *         info = 3  conditions for info = 1 and info = 2 both hold.
 * 
 *         info = 4  fvec is orthogonal to the columns of the
 *                   jacobian to machine precision.
 * 
 *         info = 5  number of calls to fcn has reached or
 *                   exceeded nsteps*(n+1).
 * 
 *         info = 6  tol is too small. no further reduction in
 *                   the sum of squares is possible.
 * 
 *         info = 7  tol is too small. no further improvement in
 *                   the approximate solution x is possible.
 * 
 *       iwa is an integer work array of length n.
 * 
 *       wa is a work array of length lwa.
 * 
 *       lwa is a positive integer input variable not less than
 *         m*n+5*n+m.
 * 
 *     subprograms called
 * 
 *       user-supplied ...... fcn
 * 
 *       minpack-supplied ... lmdif
 */


    /* Fortran parameter adjustments */
    --fvec;
    --iwa;
    --x;
    --wa;


    info = 0;

/*
 *     Check the input parameters for errors.
 */

    if (n <= 0 || m < n || *tol < zero || lwa < m * n + n * 5 + m) {
	return (int) info;
    }

    maxfev = (n + 1) * nsteps;
    
    ftol = *tol;
    xtol = zero;
    gtol = zero;
    epsfcn = zero;
    mode = 1;
    mp5n = m + n * 5;

    niter = 0;

    if (nprint != 0) {
      stufftext ("\n Trace iterations\n");
      trace_on = TRUE;
    } else  {
      trace_on = FALSE;;
    }

    lmdif_((U_fp)fcn, &m, &n, &x[1], &fvec[1], &ftol, &xtol, &gtol, &maxfev, &
	    epsfcn, &wa[1], &mode, &factor, &nprint, &info, &nfev, &wa[mp5n + 
	    1], &m, &iwa[1], &wa[n + 1], &wa[(n << 1) + 1], &wa[n * 3 + 1], 
	    &wa[(n << 2) + 1], &wa[n * 5 + 1]);
	    
    if (info == 8) {
	info = 4;
    }
    
    return (int) info;
}


/***************** Loess interface ******************/

/**
 * Assume that destination gno  gno2 == src_gno
 */
int nonlfit_loess_comand (int gno1 ,int set1
			  ,NL_family nonl_family ,int nonl_deg ,double nonl_span
			  ,double *warray
			  ,int    load_residuals
			  ,int 	  nonl_confidence_compute 
			  ,double nonl_confidence_level
			  ,int 	  nonl_confidence_len
			  ,int    gno2 ,int set2 ,int *bar_setno)
{

#ifdef HAVE_BLAS

  double *x2 ,*y2 ,*x3 ,*y3 ,*dy1 ,*dy2;
  loess_typ lo;
  char buf[64] ,*bufr;
  int i ,j ,nlen ,imin ,imax;
  double xmin ,xmax ,sm ,di;
  pred_typ     pred;
  ci_typ       ci;
  src_gno    = gno1;
  src_setno  = set1;
  dest_gno   = gno2;
  if (is_valid_setno (gno2 ,set2)) {
    dest_setno = set2;
  } else {
    dest_setno = nextset (gno2);
  }
  nlen = getsetlength (src_gno, src_setno);
  copysetdata (src_gno ,src_setno ,dest_gno ,dest_setno);
  sortset (dest_gno ,dest_setno ,DATA_X ,0);
  x2 = getx (dest_gno ,dest_setno);
  y2 = gety (dest_gno ,dest_setno);
  /* initialize loess */
  loess_setup (x2 ,y2 ,nlen ,1 ,&lo);
  lo.in.weights    = warray;
  lo.model.family  = nonl_family;
  lo.model.degree  = nonl_deg;
  lo.model.span    = nonl_span;
  
  /* compute LOESS approximation */
  if (loess (&lo) == RETURN_FAILURE) return RETURN_FAILURE;
  if (load_residuals) {
    bufr = "residuals";
    for (i = 0; i < nlen; i++) {
      y2[i] = lo.out.fitted_residuals[i];
    }
  } else {
    bufr = "fit";
    predict (x2 ,nlen ,&lo ,&pred ,FALSE);
    for (i = 0; i < nlen; i++) y2[i] = pred.fit[i];
  }
  sprintf(buf ,"LOESS %s of G%d.S%d  ->  G%d.S%d" ,bufr ,src_gno ,src_setno ,dest_setno ,dest_setno);
  comment_and_legend_set (dest_gno ,dest_setno ,buf);
  /* print summary */
  stufftext ("\n");
  stufftext (buf);
  sprintf (buf ,"\n\t Noise: %s\n" ,family[nonl_family]);
  stufftext (buf);
  sprintf (buf ,"\t degree: %s\n" ,linquad[nonl_deg-1]);
  stufftext (buf);
  sprintf (buf ,"\t smoothing span: %f\n" ,nonl_span);
  stufftext (buf);
  /*  loess_summary (&lo); */
  sprintf (buf ,"\t => Equivalent Number of Parameters: %.1f\n" ,lo.out.enp);
  stufftext (buf);
  if(nonl_family == NL_GAUSSIAN) {
    sprintf(buf ,"\t => Residual Standard Error: %.4f\n" ,lo.out.s);
  } else {
    sprintf(buf ,"\t => Residual Scale Estimate: %.4f\n" ,lo.out.s);
  }
  stufftext (buf);
  if (nonl_confidence_compute == TRUE) {
    /* compute confidence bars */
    plotarr p;
    if (nonl_confidence_len < 2) {
      errmsg ("gg_nonl_loess: number of confidence bar < 2");
      return RETURN_FAILURE;
    } else if (nonl_confidence_len > 100) {
      if (yesno_i (buf ,"number of confidence bar %d > 100: Ok to proceed? " 
		   ,nonl_confidence_len) == 0) {
	return RETURN_FAILURE;
      }
    }
    *bar_setno = nextset (dest_gno);
    setlength 	     (dest_gno ,*bar_setno ,nonl_confidence_len);
    set_dataset_type (dest_gno ,*bar_setno ,SET_XYDYDY);
    x3  = getcol     (dest_gno ,*bar_setno ,DATA_X);
    minmax 	     (x2 ,nlen ,&xmin ,&xmax ,&imin ,&imax);
    sm = (xmax + xmin)/2;
    di = (xmax - xmin)/2;
    for (i = 0; i < nonl_confidence_len; i++) {
      x3[i] = sm + di*((double) (2*i + 1 - nonl_confidence_len)/(nonl_confidence_len - 1));
    }
    predict (x3 ,nonl_confidence_len ,&lo ,&pred ,TRUE);
    pointwise (&pred ,nonl_confidence_len ,nonl_confidence_level ,&ci);
    y3  = getcol     (dest_gno ,*bar_setno ,DATA_Y);
    dy1 = getcol     (dest_gno ,*bar_setno ,DATA_Y1);
    dy2 = getcol     (dest_gno ,*bar_setno ,DATA_Y2);
    for (j = 0; j < nonl_confidence_len; j++) {
      y3[j]  = pred.fit[j];
      dy1[j] = ci.lower[j] - ci.fit[j];
      dy2[j] = ci.upper[j] - ci.fit[j];
    }
    pw_free_mem (&ci);
    sprintf (buf ,"%.3f confidence bar for G%d.S%d" ,nonl_confidence_level ,dest_setno ,dest_setno);
    comment_and_legend_set (dest_gno ,*bar_setno ,buf);
    get_graph_plotarr (dest_gno ,*bar_setno ,&p);
    p.linet = 0;  			 /* no line connection */
    p.sym = 3; 	  			 /* diamond */
    set_graph_plotarr (dest_gno ,*bar_setno ,&p);
  }
  loess_free_mem (&lo);
  if (load_residuals == FALSE) pred_free_mem (&pred);
  return RETURN_SUCCESS;

#else /* HAVE_BLAS not found => ../loess not compiled => dummy interface */
  errmsg ("GraceGTK compiled without BLAS library => LOESS not available");
  return RETURN_FAILURE;
#endif /* HAVE_BLAS */
}

/* ------------------------ Utility ------------------------ */

/**
 * Allocate and compute the weight array warray
 */
int compute_weights (int src_gno ,int src_setno
		     ,int weight_method ,const char *nonl_weight_formula
		     ,double **warray)
{
  int i ,wlen;
  double *ytmp;
  int nlen = getsetlength (src_gno, src_setno);
  if (weight_method == NL_WEIGHT_CUSTOM) {
    /* WEIGHT_CUSTOM special because warray is allocated by v_scanner */
    if (set_parser_setno (src_gno, src_setno) != RETURN_SUCCESS) {
      errmsg ("Nonl fit: bad set");
      return RETURN_FAILURE;
    }
    if (v_scanner (nonl_weight_formula, &wlen, warray) != RETURN_SUCCESS) {
      errmsg ("Nonl fit: error evaluating expression for weights");
      return RETURN_FAILURE;
    }
    if (wlen != nlen) {
      errmsg ("Nonl fit: the array of weights has different length");
      XCFREE (*warray);
      return RETURN_FAILURE;
    }
  } else {
    *warray = xmalloc (nlen*SIZEOF_DOUBLE);
    if (*warray == NULL) {
      errmsg ("Nonl fit: xmalloc failed in compute_weights()");
      return RETURN_FAILURE;
    }
    switch (weight_method) {
    case NL_WEIGHT_NONE:
      for (i = 0; i < nlen; i++) (*warray)[i] = 1.0;
      break;
    case NL_WEIGHT_Y:
    case NL_WEIGHT_Y2:                /* Y2 for y**2 */
      ytmp = getcol (src_gno, src_setno, DATA_Y);
      for (i = 0; i < nlen; i++) {
	if (ytmp[i] == 0.0) {
	  errmsg ("Nonl fit: divide by zero while calculating weights");
	  XCFREE (*warray);
	  return RETURN_FAILURE;
	}
	if (weight_method == NL_WEIGHT_Y) {
	  (*warray)[i] = 1/ytmp[i];
	} else {
	  (*warray)[i] = 1/(ytmp[i]*ytmp[i]);
	}
      }
      break;
    case NL_WEIGHT_DY:
      ytmp = getcol (src_gno, src_setno, DATA_Y1);
      if (ytmp == NULL) {
	errmsg("Nonl fit: The set doesn't have dY data column");
	return RETURN_FAILURE;
      }
      for (i = 0; i < nlen; i++) {
	if (ytmp[i] == 0.0) {
	  errmsg("Nonl fit: Divide by zero while calculating weights");
	  XCFREE (*warray);
	  return RETURN_FAILURE;
	}
	(*warray)[i] = 1/(ytmp[i]*ytmp[i]);
      }
      break;
    default:
      *warray = NULL;
      break;
    }
  }
  return RETURN_SUCCESS;
}
